/////////////////////////////////////////////////////////////////////////////
// GliD3D_Render.cpp : No comment
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// Directive

#include "GliD3D.h"
#include "gl2ideal.h"

/////////////////////////////////////////////////////////////////////////////

int			static_color_mode;
GL2IDE_COLOR	static_color;

GLsizei ArrayCount = 0;
GL2IDE_TEXCORD TexCoordArray[MAXVERTEX];
GL2IDE_COLOR ColorArray[MAXVERTEX];
CORD CoordArray[MAXVERTEX];

static GrCmpFnc_t f_alpha_test_function;
static GrAlpha_t f_alpha_test_reference_value;
static GrAlphaBlendFnc_t f_alpha_blend_function_rgb_sf;
static GrAlphaBlendFnc_t f_alpha_blend_function_rgb_df;
static FxBool f_depth_mask;
static GrCmpFnc_t f_depth_buffer_function;
static GrCullMode_t f_cull_mode;

extern GLenum	src_blendmap[16], dst_blendmap[16];
inline int FlushAlphaBlendFunction(void)
{
	if (f_alpha_blend_function_rgb_sf != g_gr_state.alpha_blend_function_rgb_sf ||
		f_alpha_blend_function_rgb_df != g_gr_state.alpha_blend_function_rgb_df){
		glBlendFunc (src_blendmap[g_gr_state.alpha_blend_function_rgb_sf], 
					dst_blendmap[g_gr_state.alpha_blend_function_rgb_df]);
		f_alpha_blend_function_rgb_sf = g_gr_state.alpha_blend_function_rgb_sf;
		f_alpha_blend_function_rgb_df = g_gr_state.alpha_blend_function_rgb_df;
		return 1;
	}
	return 0;
}

inline int FlushAlphaTestFunction(void)
{
	if(f_alpha_test_function == g_gr_state.alpha_test_function &&
		f_alpha_test_reference_value == g_gr_state.alpha_test_reference_value)
		return 0;

	f_alpha_test_function = g_gr_state.alpha_test_function;
	f_alpha_test_reference_value = g_gr_state.alpha_test_reference_value;

	glAlphaFunc (GL_NEVER + g_gr_state.alpha_test_function, DIV255(g_gr_state.alpha_test_reference_value));
	return 1;
}

inline int FlushDepthTestFunction(void)
{
	if (f_depth_buffer_function != g_gr_state.depth_buffer_function){
		f_depth_buffer_function = g_gr_state.depth_buffer_function;
		glDepthFunc (GL_NEVER + g_gr_state.depth_buffer_function);
		return 1;
	}
	return 0;
}

inline int FlushCullMode(void)
{
	if (f_cull_mode == g_gr_state.cull_mode)
		return 0;
	f_cull_mode = g_gr_state.cull_mode;

	switch (g_gr_state.cull_mode) {
	case GR_CULL_DISABLE:
		glDisable (GL_CULL_FACE);
		break;
	case GR_CULL_NEGATIVE:
		glCullFace (GL_BACK);
		glEnable (GL_CULL_FACE);
		break;
	case GR_CULL_POSITIVE:
		glCullFace (GL_FRONT);
		glEnable (GL_CULL_FACE);
		break;
	}
	return 1;
}

inline int FlushDrawState(void)
{
	grMegaCombineEverythingSetup ();

	if ((g_gr_state.alpha_blend_function_rgb_sf == GR_BLEND_ONE &&
		 g_gr_state.alpha_blend_function_rgb_df == GR_BLEND_ZERO)){
		DisableBlend();
	}else{
		FlushAlphaBlendFunction();
		EnableBlend();
	}
	if (g_gr_state.alpha_combine_other != GR_COMBINE_OTHER_TEXTURE)
		DisableAlphaTest();
	else{
		FlushAlphaTestFunction();
		EnableAlphaTest();
	}
	if (f_depth_mask != g_gr_state.depth_mask){
		f_depth_mask = g_gr_state.depth_mask;
		glDepthMask (g_gr_state.depth_mask);
	}
	FlushDepthTestFunction();
	FlushCullMode();

	return 1;
}

int InitDrawState(void)
{
	f_alpha_test_function = g_gr_state.alpha_test_function;
	f_alpha_test_reference_value = g_gr_state.alpha_test_reference_value;
	f_alpha_blend_function_rgb_sf = g_gr_state.alpha_blend_function_rgb_sf;
	f_alpha_blend_function_rgb_df = g_gr_state.alpha_blend_function_rgb_df;
	f_depth_mask = g_gr_state.depth_mask;
	f_depth_buffer_function = g_gr_state.depth_buffer_function;
	f_cull_mode = g_gr_state.cull_mode;
	return 1;
}

/////////////////////////////////////////////////////////////////////////////
// Defines / Variables

#define GLID3D_ANTIALIAS_ENABLE
#define GLID3D_ANTIALIAS_DISABLE

/////////////////////////////////////////////////////////////////////////////
// Prototypes

/////////////////////////////////////////////////////////////////////////////
// Implements

FX_ENTRY void FX_CALL grDrawLine (const GrVertex * a, const GrVertex * b)
{
    VERTEX2 ptlv[2];

    TRACEFUNCTION ("grDrawLine");

	FlushTriangleVertex();

    grsubSetVertexFromGrVertex (ptlv[0].cord, ptlv[0].color, ptlv[0].tcord, a);
	 grsubSetVertexFromGrVertex (ptlv[1].cord, ptlv[1].color, ptlv[1].tcord, b);
    glBegin (GL_LINES);
    glTexCoord2fv ((const GLfloat *)&ptlv[0].tcord);
    glColor4fv ((const GLfloat *)&ptlv[0].color);
    glVertex3fv ((const GLfloat *)&ptlv[0].cord);
    glTexCoord2fv ((const GLfloat *)&ptlv[1].tcord);
    glColor4fv ((const GLfloat *)&ptlv[1].color);
    glVertex3fv ((const GLfloat *)&ptlv[1].cord);
    glEnd ();

}

FX_ENTRY void FX_CALL grDrawPlanarPolygon (int nverts, const int ilist[], const
    GrVertex vlist[])
{
    TRACEFUNCTION ("grDrawPlannarPolygon");

    grDrawPolygon (nverts, ilist, vlist);
}

FX_ENTRY void FX_CALL grDrawPlanarPolygonVertexList (int nverts, const
    GrVertex vlist[])
{
    TRACEFUNCTION ("grPlanarPolygonVertexList");

    grDrawPolygonVertexList (nverts, vlist);
}

FX_ENTRY void FX_CALL grDrawPoint (const GrVertex * pt)
{
    VERTEX2 ptlv[1];
    TRACEFUNCTION ("grDrawPoint");

	FlushTriangleVertex();
    grsubSetVertexFromGrVertex (ptlv[0].cord, ptlv[0].color, ptlv[0].tcord, pt);
    glBegin (GL_POINTS);
    glColor4fv ((const GLfloat *)&ptlv[0].color);
    glVertex3fv ((const GLfloat *)&ptlv[0].cord);
    glEnd ();
}

FX_ENTRY void FX_CALL grDrawPolygon (int nverts, const int ilist[], const
    GrVertex vlist[])
{
    TRACEFUNCTION ("grDrawPolygon");

    switch (g_gr_state.origin_location)
    {
    case GR_ORIGIN_UPPER_LEFT:
     {
         break;
     }
    case GR_ORIGIN_LOWER_LEFT:
     {
         break;
     }
    }

}

// Untested
FX_ENTRY void FX_CALL grDrawPolygonVertexList (int nverts, const GrVertex
    vlist[])
{
    static VERTEX2 ptlv;
    register int i;

    TRACEFUNCTION ("grDrawPolygonVertexList");

	FlushTriangleVertex();

    glBegin (GL_POLYGON);
    if (g_gr_state.origin_location==GR_ORIGIN_UPPER_LEFT)
    {
       for (i = 0; i < nverts; i++)
       {
             grsubSetVertexFromGrVertex (ptlv.cord, ptlv.color, ptlv.tcord, &vlist[i]);
             glTexCoord2fv ((const GLfloat *)&ptlv.tcord);
             glColor4fv ((const GLfloat *)&ptlv.color);
             glVertex4fv ((const GLfloat *)&ptlv.cord);
       }
    } else
    {
       for (i = nverts - 1; i >= 0; i--)
       {
             grsubSetVertexFromGrVertex (ptlv.cord, ptlv.color, ptlv.tcord, &vlist[i]);
             glTexCoord2fv ((const GLfloat *)&ptlv.tcord);
             glColor4fv ((const GLfloat *)&ptlv.color);
             glVertex4fv ((const GLfloat *)&ptlv.cord);
       }
    }
    glEnd ();
}


FX_ENTRY void FX_CALL grDrawTriangle (const GrVertex * a, const GrVertex *
	 b, const GrVertex * c)
{
	 TRACEFUNCTION ("grDrawTriangle");
    if (!g_gr_DrawingTriangles)
    {
       g_gr_DrawingTriangles = TRUE;
       ArrayCount=0;
       FlushDrawState();
    }
	if (g_gr_state.origin_location==GR_ORIGIN_UPPER_LEFT) {
	    grsubSetVertexFromGrVertex(CoordArray[ArrayCount],
	      ColorArray[ArrayCount], TexCoordArray[ArrayCount], c);
	    ArrayCount++;
	    grsubSetVertexFromGrVertex(CoordArray[ArrayCount],
	      ColorArray[ArrayCount], TexCoordArray[ArrayCount], b);
	    ArrayCount++;
	    grsubSetVertexFromGrVertex(CoordArray[ArrayCount],
	      ColorArray[ArrayCount], TexCoordArray[ArrayCount], a);
	    ArrayCount++;
	} else {
	    grsubSetVertexFromGrVertex(CoordArray[ArrayCount],
	      ColorArray[ArrayCount], TexCoordArray[ArrayCount], a);
	    ArrayCount++;
	    grsubSetVertexFromGrVertex(CoordArray[ArrayCount],
	      ColorArray[ArrayCount], TexCoordArray[ArrayCount], b);
	    ArrayCount++;
	    grsubSetVertexFromGrVertex(CoordArray[ArrayCount],
	      ColorArray[ArrayCount], TexCoordArray[ArrayCount], c);
	    ArrayCount++;
	}
	if(ArrayCount == 3){
		if(	g_gr_state.color_combine_factor == GR_COMBINE_FACTOR_TEXTURE_RGB &&
			g_gr_state.color_combine_local == GR_COMBINE_LOCAL_CONSTANT &&
			g_gr_state.color_combine_other == GR_COMBINE_OTHER_ITERATED)
		{
			const GL2IDE_COLOR	color = { a->r/255.0f, a->g/255.0f, a->b/255.0f, a->a/255.0f};
			glTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, &color.r);
			FlushTriangleVertex();
		}else
		if(	!g_gr_bUseTexEnvCombine &&
			g_gr_state.color_combine_factor == GR_COMBINE_FACTOR_ONE &&
			g_gr_state.color_combine_function == GR_COMBINE_FUNCTION_LOCAL &&
		//	g_gr_state.color_combine_local == GR_COMBINE_LOCAL_ITERATED &&
		//	g_gr_state.color_combine_other == GR_COMBINE_OTHER_CONSTANT &&
			g_gr_state.alpha_combine_function == GR_COMBINE_FUNCTION_SCALE_OTHER &&
			g_gr_state.alpha_combine_other == GR_COMBINE_OTHER_TEXTURE)
		{
			glTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, &ColorArray[0].r);
			FlushTriangleVertex();
		}
	}
	if (ArrayCount==MAXVERTEX)
		FlushTriangleVertex();
}

FX_ENTRY void FX_CALL guAADrawTriangleWithClip (const GrVertex * a, const GrVertex
	 * b, const GrVertex * c)
{
	 TRACEFUNCTION ("guAADrawTriangleWithClip");

	 GLID3D_ANTIALIAS_ENABLE
	 guDrawTriangleWithClip (a, b, c);
	 GLID3D_ANTIALIAS_DISABLE
}

FX_ENTRY void FX_CALL guDrawTriangleWithClip (const GrVertex * a,
	 const GrVertex * b,
	 const GrVertex * c
)
{
	 TRACEFUNCTION ("guDrawTriangleWithClip");
	 grDrawTriangle (a, b, c);
}

FX_ENTRY void FX_CALL guDrawPolygonVertexListWithClip (int nverts, const
	 GrVertex vlist[])
{
	 TRACEFUNCTION ("guDrawPolygonVertexListWithClip");
	 grDrawPolygonVertexList (nverts, vlist);
}

/////////////////////////////////////////////////////////////////////////////


