
//#define USE_VFUNC_LUT

#ifndef GL_EXT_texture_env_combine
#define GL_COMBINE_EXT			0x8570
#define GL_COMBINE_RGB_EXT		0x8571
#define GL_COMBINE_ALPHA_EXT	0x8572
#define GL_SOURCE0_RGB_EXT		0x8580
#define GL_SOURCE0_ALPHA_EXT	0x8588
#endif	//GL_EXT_texture_env_combine

#ifndef GL_PRIMARY_COLOR_EXT
#define GL_PRIMARY_COLOR_EXT	0x8577
#endif

#define TEXID_INVALID	0xffffffff
#define DIV255(x) div255_lookuptable[x]
#define STATIC_COLOR_MODE(RGB_IT, ALPHA_IT) ((RGB_IT == GR_COMBINE_LOCAL_ITERATED) << 1 | (ALPHA_IT == GR_COMBINE_LOCAL_ITERATED))

typedef VOID (*vertexconv_func) (CORD &cord, GL2IDE_COLOR &color,
       GL2IDE_TEXCORD &texcord, const GrVertex * pv);

extern int			static_color_mode;
extern GL2IDE_COLOR	static_color;

extern GLfloat	div255_lookuptable[256];

extern BOOL g_gr_AlphaTestOn;
extern BOOL g_gr_BlendOn;
extern BOOL g_gr_Texture2DOn;
extern BOOL g_gr_FogOn;
extern GLint g_gr_texenv;

void InitBlendMap(void);
void InitDiv255LookupTable(void);

HWND
CreateOpenGLWindow (HWND Parent_hWnd, char *title, int x, int y, int width, int height,
    BYTE type, DWORD flags);

int InitDrawState(void);
void InitGLExt(void);

/////////////////////////////////////////////////////////////////////////////

#ifdef USE_VFUNC_LUT
void InitLookupVertexFuncs (void);
void ChooseVertexFunc (void);
extern vertexconv_func grsubSetVertexFromGrVertex;
#else
inline void InitLookupVertexFuncs (void){}
inline void ChooseVertexFunc (void){}
inline void grsubSetVertexFromGrVertex (CORD &cord, GL2IDE_COLOR &color,
       GL2IDE_TEXCORD &tcord, const GrVertex * pv)
{
	const int nTmu = 0;
	const GLfloat f_255 = 1.0/255.0;
	const GLfloat f4096 = 4096.0;

	if(g_gr_state.origin_location == GR_ORIGIN_UPPER_LEFT){
		if (pv->y < f4096)
			cord.y = pv->y;
		else
			cord.y = pv->y - (FLOAT) (((DWORD) pv->y) & 0xfffff000);
	}else{
		if (pv->y < f4096)
			cord.y = (FLOAT) g_gr_nHeight - pv->y;
		else
			cord.y = (FLOAT) g_gr_nHeight - (FLOAT) (((DWORD) pv->y) & 0xfffff000);
	}

	if (pv->x < f4096)
		cord.x = pv->x;
	else
		cord.x = pv->x - (FLOAT) (((DWORD) pv->x) & 0xfffff000);
	cord.w = 1.0 / pv->oow;
		 // scale texture coordinates according to aspect ratio
	if(g_gr_Texture2DOn){
		tcord.tu = pv->tmuvtx[0].sow * cord.w * g_gr_tmu[nTmu].ah;
		tcord.tv = pv->tmuvtx[0].tow * cord.w * g_gr_tmu[nTmu].aw;
	}

	if (g_gr_bUseZInverse) {
		if (g_gr_bPrecisionFix) {
			if (cord.w < 10)
			cord.w = 10;
		}
		cord.z = 1.0f;
	}else{
		cord.z = -cord.w * cord.w;
	}
	cord.x *= cord.w;
	cord.y *= cord.w;

	switch(static_color_mode){
	case STATIC_COLOR_MODE(GR_COMBINE_LOCAL_CONSTANT,GR_COMBINE_LOCAL_CONSTANT):
		color = static_color;
		break;
	case STATIC_COLOR_MODE(GR_COMBINE_LOCAL_CONSTANT,GR_COMBINE_LOCAL_ITERATED):
		color = static_color;
		color.a = pv->a * f_255;
		break;
	case STATIC_COLOR_MODE(GR_COMBINE_LOCAL_ITERATED,GR_COMBINE_LOCAL_CONSTANT):
		color.r = pv->r * f_255;
		color.g = pv->g * f_255;
		color.b = pv->b * f_255;
		color.a = static_color.a;
		break;
	default:
		color.r = pv->r * f_255;
		color.g = pv->g * f_255;
		color.b = pv->b * f_255;
		color.a = pv->a * f_255;
		break;
	}
}
#endif	//USE_VFUNC_LUT

inline void EnableBlend(void)
{
   if (!g_gr_BlendOn)
   {
       glEnable(GL_BLEND);
       g_gr_BlendOn = TRUE;
   }
}

inline void DisableBlend(void)
{
   if (g_gr_BlendOn)
   {
      glDisable(GL_BLEND);
      g_gr_BlendOn = FALSE;
   }
}

inline static void EnableAlphaTest(void)
{
   if (!g_gr_AlphaTestOn)
   {
       glEnable(GL_ALPHA_TEST);
       g_gr_AlphaTestOn = TRUE;
   }
}

inline static void DisableAlphaTest(void)
{
   if (g_gr_AlphaTestOn)
   {
      glDisable(GL_ALPHA_TEST);
      g_gr_AlphaTestOn = FALSE;
   }
}

inline void EnableTexture2D(void)
{
   if (!g_gr_Texture2DOn)
   {
       glEnable(GL_TEXTURE_2D);
       g_gr_Texture2DOn = TRUE;
   }
}

inline void DisableTexture2D(void)
{
   if (g_gr_Texture2DOn)
   {
      glDisable(GL_TEXTURE_2D);
	   g_gr_Texture2DOn = FALSE;
   }
}

inline void EnableFog(void)
{
   if (!g_gr_FogOn)
   {
       glEnable(GL_FOG);
       g_gr_FogOn = TRUE;
   }
}

inline void DisableFog(void)
{
   if (g_gr_FogOn)
   {
      glDisable(GL_FOG);
      g_gr_FogOn = FALSE;
   }
}

inline void SetglTexEnvi(GLenum target, GLenum pname, GLint texenv)
{
	if(g_gr_texenv != texenv){
		g_gr_texenv = texenv;
		glTexEnvi(target, pname, texenv);
	}
}

inline void grMegaCombineEverythingSetup ()
{
	static_color = g_gr_state.constant_color;
	static_color_mode = STATIC_COLOR_MODE(g_gr_state.color_combine_local, g_gr_state.alpha_combine_local);

	if((g_gr_state.color_combine_function == GR_COMBINE_FUNCTION_ZERO ||
		g_gr_state.color_combine_factor == GR_COMBINE_FACTOR_NONE ||
		(g_gr_state.color_combine_factor != GR_COMBINE_FACTOR_TEXTURE_RGB &&
		g_gr_state.color_combine_other != GR_COMBINE_OTHER_TEXTURE)) &&

		(g_gr_state.alpha_combine_function == GR_COMBINE_FUNCTION_ZERO ||
		g_gr_state.alpha_combine_factor == GR_COMBINE_FACTOR_NONE ||
		(g_gr_state.alpha_combine_factor != GR_COMBINE_FACTOR_TEXTURE_RGB &&
		g_gr_state.alpha_combine_other != GR_COMBINE_OTHER_TEXTURE)))
	{
		DisableTexture2D();
		ChooseVertexFunc();
	}else{

		if (g_gr_state.color_combine_function == GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL &&
			g_gr_state.color_combine_factor == GR_COMBINE_FACTOR_TEXTURE_ALPHA &&
			g_gr_state.color_combine_other == GR_COMBINE_OTHER_TEXTURE &&
			g_gr_state.color_combine_local == GR_COMBINE_LOCAL_ITERATED)
		{
			SetglTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
		}else
		if(g_gr_state.color_combine_factor == GR_COMBINE_FACTOR_TEXTURE_RGB)
		{
			SetglTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_BLEND);	// or Reverse_Blend
		}else
		if (g_gr_state.color_combine_function == GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL &&
			g_gr_state.color_combine_local == GR_COMBINE_LOCAL_ITERATED &&
			g_gr_state.color_combine_factor == GR_COMBINE_FACTOR_ONE)
		{
			if(g_gr_bUseTexEnvAdd){
				SetglTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_ADD);
			}else{
				static_color.r = static_color.g = static_color.b = 1;
				static_color_mode = STATIC_COLOR_MODE(GR_COMBINE_LOCAL_CONSTANT, g_gr_state.alpha_combine_local);
				SetglTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
			}
		}else
		if(	g_gr_state.color_combine_factor == GR_COMBINE_FACTOR_ONE &&
			g_gr_state.color_combine_function == GR_COMBINE_FUNCTION_LOCAL &&
			g_gr_state.color_combine_local == GR_COMBINE_LOCAL_ITERATED &&
			g_gr_state.color_combine_other == GR_COMBINE_OTHER_CONSTANT &&
			g_gr_state.alpha_combine_function == GR_COMBINE_FUNCTION_SCALE_OTHER &&
			//g_gr_state.alpha_combine_local == GR_COMBINE_LOCAL_ITERATED &&
			g_gr_state.alpha_combine_other == GR_COMBINE_OTHER_TEXTURE)
		{
			if(g_gr_bUseTexEnvCombine){
				SetglTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB_EXT, GL_REPLACE);
				SetglTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_RGB_EXT, GL_PRIMARY_COLOR_EXT);
				SetglTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_ALPHA_EXT, GL_REPLACE);
				SetglTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_ALPHA_EXT, GL_TEXTURE);
				SetglTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_EXT);
			//	static_color_mode = STATIC_COLOR_MODE(GR_COMBINE_LOCAL_CONSTANT, g_gr_state.alpha_combine_local);
			}else{
				// Do same-color-blending to emulate this
				// It is not perfect. (when color_combine_local == GR_COMBINE_LOCAL_ITERATED ...)
				SetglTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_BLEND);	// or Reverse_Blend
			}
		}else
		if (g_gr_state.color_combine_function == GR_COMBINE_FUNCTION_SCALE_OTHER &&
			g_gr_state.color_combine_other == GR_COMBINE_OTHER_TEXTURE)
		{
#if 0
			SetglTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
#else
			if(g_gr_state.color_combine_factor == GR_COMBINE_FACTOR_ONE){
				static_color.r = static_color.g = static_color.b = 1;
			}else if(g_gr_state.color_combine_factor == GR_COMBINE_FACTOR_ZERO){
				static_color.r = static_color.g = static_color.b = 0;
			}
			//static_color_mode = STATIC_COLOR_MODE(GR_COMBINE_LOCAL_CONSTANT, g_gr_state.alpha_combine_local);
			SetglTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
#endif
		}else{
			SetglTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
		}

		if (g_gr_state.alpha_combine_function == GR_COMBINE_FUNCTION_SCALE_OTHER &&
			g_gr_state.alpha_combine_other == GR_COMBINE_OTHER_TEXTURE)
		{
			if (g_gr_state.alpha_combine_factor == GR_COMBINE_FACTOR_ONE)
				static_color.a = 1;
			else if (g_gr_state.alpha_combine_factor == GR_COMBINE_FACTOR_ZERO)
				static_color.a = 0;
			//static_color_mode &= STATIC_COLOR_MODE(GR_COMBINE_LOCAL_ITERATED, GR_COMBINE_LOCAL_CONSTANT);
		}

		EnableTexture2D();
		ChooseVertexFunc();
	}

}


/////////////////////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////////////////////

